#
#   Created by Ryan McDermott
#   Created on 3/16/2022
#

from google.cloud import bigquery
from google.api_core.exceptions import NotFound
from google.api_core.exceptions import ServiceUnavailable
from indxdatalaketools import Helpers
from indxdatalaketools.DataLakeGlobals import DATASET_OWNER
from indxdatalaketools.GoogleClients.GoogleClients import GcpClients
from indxdatalaketools.Helpers import print_error
from google.cloud.bigquery.enums import EntityTypes
from indxdatalaketools.GoogleClientWrappers.retry_decorator import retry_decorator


class Wrapper:
    '''
        Wrapper for Google Cloud Platform Big Query python SDK
    '''

    __google_big_query_client = None
    file_format_map = {
        'AVRO': bigquery.SourceFormat.AVRO,
        'CSV': bigquery.SourceFormat.CSV
    }

    def __init__(self):
        self.__google_big_query_client = GcpClients.instance(
        ).get_bigquery_client()

    @retry_decorator()
    def insert_to_bq_table(self, patient_data, table_name):
        '''
            Function that inserts data into a big query table
            Args:
                patient_data (list): a list of dictionaries containing  patient data
                table_name (string): the name of the table we want to insert to
            Returns:
                boolean: True if the data was inserted, False if otherwise
        '''
        try:
            errors = GcpClients.instance().get_bigquery_client(
            ).insert_rows_json(table_name, patient_data)
            if errors == []:
                return True
            print_error(errors)
            return False
        except ServiceUnavailable:
            raise ServiceUnavailable('503 retry')
        except Exception as e:
            print_error(str(e))
            return False

    @retry_decorator()
    def batch_load_from_gcs(self, gcs_uri, table_destination, file_format):
        '''
            Batch loads a file from gcs into big query
            Args:
                gcs_uri (string): the gcs uri of the file we wish to load into big query
                table_destination (string): The destination table
                file_format (string): the file format we are loading into big query 
                    (AVRO, CSV)
            Returns:
                boolean: True if the operation was successful, False if otherwise
        '''
        job_config = bigquery.LoadJobConfig(
            use_avro_logical_types=True,
            write_disposition=bigquery.WriteDisposition.WRITE_TRUNCATE,
            source_format=self.file_format_map[file_format])
        try:
            load_job = self.__google_big_query_client.load_table_from_uri(
                gcs_uri, table_destination,
                job_config=job_config)  # Make an API request.
            load_job.result()
            return True
        except ServiceUnavailable as e:
            raise ServiceUnavailable('503 retry')
        except Exception as e:
            print_error(str(e))
            return False

    def create_dataset_if_not_exist(self, dataset_id, location):
        '''
            Creates the client's Dataset if it does not already exist
            Args:
                dataset_id (string): The id of the dataset for the table
                location (string): The location of the dataset
            Returns:
                boolean: True if the dataset already exists or was created, False if otherwise
        '''
        dataset = self.get_dataset(dataset_id)

        if dataset is None:
            return False

        dataset.location = location
        # Check if dataset exists
        if self.check_if_dataset_exists(dataset_id):
            print("Dataset {} already exists, skipping this step".format(
                dataset_id))
            return True

        print("Creating dataset: " + dataset_id)

        return self.create_dataset(dataset)

    @retry_decorator()
    def get_dataset(self, dataset_id):
        '''
            Gets the dataset in bigquery
            Args:
                dataset_id (string): The dataset Id we wish to retrieve or create
            Returns:
                google.cloud.bigquery.Dataset
        '''
        try:
            return bigquery.Dataset(dataset_id)
        except ServiceUnavailable as e:
            raise ServiceUnavailable('503 retry')
        except Exception as e:
            Helpers.print_error('Could not retreive dataset ' + dataset_id +
                                " " + str(e))
            return None

    def create_table_if_not_exist(self,
                                  dataset_id,
                                  table_name,
                                  schema,
                                  partition=None,
                                  clustering=None):
        '''Creates the client's PATIENT table if it does not already exist

        This function also takes in optional partitions and clustering for the table to be
        created

        Args:
            dataset_id (string): The id of the dataset for the table
            table_name (string): The name of the table we wish to create
            schema (list): a list of bigquery.SchemaField representing the table schema
        Returns:
            boolean: True if the table was created or already existed, false
                if table creation failed
        '''
        # create table object
        table_id = dataset_id + "." + table_name
        table = bigquery.Table(table_id, schema=schema)
        if partition is not None:
            table.range_partitioning = bigquery.RangePartitioning(
                # To use integer range partitioning, select a top-level REQUIRED /
                # NULLABLE column with INTEGER / INT64 data type.
                field=partition,
                range_=bigquery.PartitionRange(start=0, end=4000, interval=1),
            )

        if clustering is not None:
            table.clustering_fields = clustering

        if self.check_if_table_exists(table_id):
            print(
                "Table {} already exists, skipping this step".format(table_id))
            return True

        print("Creating table " + table_id)
        return self.create_table(table)

    @retry_decorator()
    def create_table(self, table):
        '''
            Creates Big Query Table
            Args:
                table (bigquery.Table): The table we wish to create
            returns:
                bigquery.Table: The created Table
        '''
        try:
            self.__google_big_query_client.create_table(table)
            return True
        except ServiceUnavailable as e:
            raise ServiceUnavailable('503 retry')
        except Exception as e:
            Helpers.print_error("ERROR in creating table: " + str(e))
            return False

    @retry_decorator()
    def create_dataset(self, dataset):
        '''
            Creates Big Query Dataset
            Args:
                dataset (bigquery.Dataset): The dataset we wish to create
            returns:
                None
        '''
        try:
            self.__google_big_query_client.create_dataset(dataset, timeout=30)
            return True
        except ServiceUnavailable as e:
            raise ServiceUnavailable('503 retry')
        except Exception as e:
            Helpers.print_error("Could not connect to GCS client " + str(e))
            return False

    @retry_decorator()
    def check_if_table_exists(self, table_id):
        '''
            Checks if the table Id passed exists on GCP
            Args:
                dataset_id (string): The table {PROJECT_ID}.{CLIENT_UUID}.PATIENTS
                retries (int): Optional parameter for the number of retries this function will do
            Returns:
                boolean: True if the table exists, false if other wise
        '''
        try:
            self.__google_big_query_client.get_table(
                table_id)  # Make an API request.
            return True
        except ServiceUnavailable as e:
            raise ServiceUnavailable('503 retry')
        except NotFound:
            return False
        except Exception as e:
            print(str(e))
            return False

    @retry_decorator()
    def check_if_dataset_exists(self, dataset_id):
        '''
            Checks if the dataset ID passed exists on GCP
            Args:
                dataset_id (string): The dataset {PROJECT_ID}.{CLIENT_UUID}
            Returns:
                boolean: True if the datset exists, false if other wise
        '''
        try:
            self.__google_big_query_client.get_dataset(
                dataset_id)  # Make an API request.
            return True
        except ServiceUnavailable as e:
            raise ServiceUnavailable('503 retry')
        except NotFound:
            return False
        except Exception as e:
            print(str(e))
            return False

    @retry_decorator()
    def delete_table(self, table_id):
        '''
            Deletes a table
            Args:
                table_id (string): The table id (your-project.your_dataset.your_table)
            Returns:
                boolean: True if the table was deleted false if otherwise
        '''
        try:
            self.__google_big_query_client.delete_table(table_id)
            return True
        except ServiceUnavailable as e:
            raise ServiceUnavailable('503 retry')
        except Exception as e:
            print_error('Failed to Delete table: ' + table_id + " " + str(e))
            return False

    @retry_decorator()
    def perform_query(self, query, job_config=None):
        '''Performs a bigquery query

        An optional job config can be passed, usually used when querying clustered data

        Args:
            query (string): the query we wish to perform
            job_config(google.cloud.bigquery.JobConfig): an optional job configuration for
                the query
        Returns:
            query_job: The job of the query 
        '''
        try:
            query_job = self.__google_big_query_client.query(
                query, job_config=job_config)
            rows = query_job.result()
            return rows
        except ServiceUnavailable as e:
            raise ServiceUnavailable('503 retry')
        except Exception as e:
            Helpers.print_error('Could not connect to GCS Big Query ' + str(e))
            return None

    def set_permissions_on_dataset(self, dataset_id, role, account):
        '''
            Function that sets the permissions of a dataset
            Args:
                dataset_id (string): The dataset
                role (string): The role we want the account to have
                account (string): The account we are giving permission to
            Returns:
                boolean: True if the permissions were set, False if otherwise
        '''
        dataset = self.get_dataset(dataset_id)
        entity_type = EntityTypes.USER_BY_EMAIL
        entries = list(dataset.access_entries)
        # need to add an owner role to the dataset to set the SA permissions
        entries.append(
            bigquery.AccessEntry(
                role='roles/bigquery.dataOwner',
                entity_type=entity_type,
                entity_id=DATASET_OWNER,
            ))
        entries.append(
            bigquery.AccessEntry(
                role=role,
                entity_type=entity_type,
                entity_id=account,
            ))

        dataset.access_entries = entries

        return self.update_dataset(dataset, ["access_entries"])

    @retry_decorator()
    def update_dataset(self, dataset, fields):
        '''
            Updates a dataset with the fields passed
            Args:
                dataset (bigquery.Dataset): The dataset we wish to update
                fields (list): list of fields we wish to update
            Returns:
                boolean: True if the dataset was updated
        '''
        try:
            self.__google_big_query_client.update_dataset(dataset, fields)
            return True
        except ServiceUnavailable as e:
            raise ServiceUnavailable('503 retry')
        except Exception as e:
            print_error(e)
            return False



#
#   Created by Ryan McDermott
#   Created on 3/11/2022
#

import datetime

from google.cloud.bigquery_storage_v1 import types
from google.cloud.bigquery_storage_v1 import writer
from google.protobuf import descriptor_pb2

from indxdatalaketools import Helpers
from indxdatalaketools.GoogleClientWrappers import BigQuery
from indxdatalaketools.GoogleClients.GoogleClients import GcpClients


class Client:
    ''' Class that inserts data into the changelog table '''
    client_uuid = ''
    table_fields = [
        "CLIENT_ID", "BUCKET_ID", "FILE_PATH", "FILE_NAME", "ACTION",
        "ACTION_AT", "REPLICATED", "REPLICATED_AT"
    ]

    __big_query_wrapper = None

    def __init__(self, client_uuid):
        '''
            Init function that sets up client credentials, uuid, 
            and google api clients used to insert items to the Patients table
            Args:
                google_clients          (ApiClients): object containing all Google Api Clients
                client_uuid             (string): The client's uuid
            Returns:
                None
        '''
        self.client_uuid = client_uuid
        logging_client = GcpClients.instance().get_logging_client()
        self.__changelog_insert_logger = logging_client.logger(
            client_uuid + '-changelog-logger')
        self.__big_query_wrapper = BigQuery.Wrapper()

    def insert_to_changelog_table(self, blob, action):
        ''' 
            Inserts information from an uploaded blob to the changelog table
            Args:
                blob (google.cloud.storage.Blob): The blob that was uploaded
            Returns:
                boolean: True if the data was inserted into the changelog table
        '''
        blob_data_dict = self.__create_upload_dict(blob, action)
        dataset_id = 'SNAPSHOTS_' + self.__transform_client_id()
        today = self.__get_date_today()

        return self.__insert_data_to_changelog(dataset_id, today,
                                               blob_data_dict)

    def __create_upload_dict(self, blob, action):
        '''
            Creates a dictionary with all of the data for the table
            Args:
                blob   (google.cloud.storage.Blob): The blob that was 
                    created, updated, or deleted
                action (string): the action being taken
            Returns:
                dict: The dictionary containing all of the data
        '''
        time_now = self.__get_datetime_now()

        client_id = self.client_uuid
        bucket_id = self.client_uuid
        file_path = blob.metadata['FILE_PATH']
        file_name = str(blob.name).split('/')[-1]
        action_at = time_now
        replicated = False

        return {
            'CLIENT_ID': client_id,
            'BUCKET_ID': bucket_id,
            'FILE_PATH': file_path,
            'FILE_NAME': file_name,
            'ACTION': action,
            'ACTION_AT': action_at,
            'REPLICATED': replicated
        }

    def __get_datetime_now(self):
        '''
            Gets the current datetime for now and returns it
            Returns:
                datetime: The datetime of now
        '''
        now = datetime.datetime.utcnow()
        time_now_string = now.strftime('%Y-%m-%d %H:%M:%S')

        return time_now_string

    def __transform_client_id(self):
        '''
            Transforms the client id to contain no dashes and have all
            letters uppercase:
            Args:
                None
            Returns:
                string: transformed string
        '''
        client_id = self.client_uuid
        client_id = client_id.replace('-', '')
        client_id = client_id.upper()

        return client_id

    def __get_date_today(self):
        '''
            creates a date string in YYYYMMDD format by using the date today 
            Returns:
                string: Date today in YYYYMMDD format
        '''
        today = datetime.datetime.utcnow()
        today_date = today.date()
        today = today_date.strftime('%Y%m%d')

        return str(today)

    def __insert_data_to_changelog(self, dataset_id, table_name,
                                   changelog_data):
        '''
            Function that inserts data into the changelog table
            Args:
                dataset_id        (string): The id of the data set
                table_name        (string): The name of the table
                changelog_data    (dict): The Patients Data
            Returns
                Boolean: True if the request worked, false if other wise
        '''
        return self.__big_query_wrapper.insert_to_bq_table(
            [changelog_data],
            'indx-data-services.' + dataset_id + '.' + table_name)



#
#   Created by Ryan McDermott
#   Created on  2/24/2022
#

import re
import cv2
import os

from indxdatalaketools.GoogleClients.GoogleClients import GcpClients

import numpy as np

from indxdatalaketools import Helpers
from indxdatalaketools.ClientTools.PatientsTable import Insert
from indxdatalaketools.ClientTools.Changelog import Insert as ChnglogInsert
from indxdatalaketools.DataStandardizer import PatientTable

from indxdatalaketools import DataLakeGlobals


class Client:
    '''
        Class that contains all the functionality to upload a file
        to a data lake
    '''

    client_uuid = ''
    text_file_list = [".txt", ".tsv", ".csv"]

    def __init__(self, client_uuid):
        '''
            Init function that sets up client credentials, uuid, 
            and google api clients used to upload files to a GCS bucket
            Args:
                google_clients          (ApiClients): object containing all Google Api Clients
                client_uuid             (string): The client's uuid
            Returns:
                None
        '''
        self.client_uuid = client_uuid
        self.__file_upload_logger = GcpClients.instance().get_logging_client(
        ).logger(client_uuid + '-file-logger')

    def upload_file_to_datalake(self, modality, mrn, metadata, file_path):
        '''
            Uploads a file to the data lake and sets the metadata
            for that file
            Args:
                modality    (string): The file's modality
                mrn         (string): The mrn of the patient to who the files belongs to
                metadata    (string): The metadata that is associated with the
                    file
                file_path   (string): The path to the file we wish to upload to a 
                    datalake
            Returns:
                boolean: True if the File was successfully uploaded, False if 
                    otherwise
                
        '''
        file_contents = self.__read_file_contents(file_path)
        file_extension = os.path.splitext(file_path)[1]

        DataLakeGlobals.SIZE_OF_FILES += len(file_contents) * .000001

        # convert bmp to png
        if '.bmp' == file_extension.lower():
            file_contents = self.__convert_bmp_to_png(file_contents)
            file_root = os.path.splitext(file_path)[0]
            file_path = file_root + '.png'

        stripped_mrn = self.__strip_leading_zeros(mrn)
        hashed_patient_mrn = Helpers.patient_hash(self.client_uuid,
                                                  [stripped_mrn])[0]
        hashed_file_name = Helpers.hash_contents_to_file_name(
            file_path, file_contents)
        blob_name = modality + "/" + hashed_patient_mrn + "/" + hashed_file_name
        bucket = GcpClients.instance().get_storage_client().bucket(
            self.client_uuid)
        blob = bucket.blob(blob_name)

        if '.pdf' == file_extension.lower():
            blob.content_type = "application/pdf"

        if Helpers.try_blob_exists(blob):
            Helpers.print_error("File" + blob.name + " already exists")
            self.__file_upload_logger.log_text("File" + blob.name +
                                               " already exists",
                                               severity='INFO')
            return True

        if not self.__create_blob_with_metadata(blob, file_contents, metadata):
            Helpers.print_error("Could not upload file " + file_path +
                                " to datalake")
            return False

        if hashed_patient_mrn != 'UNKNOWN'and \
            not self.__insert_data_to_patients_table(stripped_mrn, hashed_patient_mrn):
            Helpers.print_error("Could not insert data to PATIENTS table")
            return False

        blob.metadata = metadata
        return ChnglogInsert.Client(self.client_uuid).insert_to_changelog_table(
            blob, 'CREATED')

    def __convert_bmp_to_png(self, file_bytes):
        '''
            Converts the BMP bytes string into a png bytes string
            Args:
                byte (string): The string of bytes we wish to convert
            Returns:
                string: The converted bytes in string format
        '''
        nparr = np.frombuffer(file_bytes, dtype=np.uint8)
        img = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
        png_img = cv2.imencode('.png', img)[1]
        data_encode = np.array(png_img)
        byte_encode = data_encode.tobytes()

        return byte_encode

    def __strip_leading_zeros(self, string_to_strip):
        '''
            strips the leading '0' from a string
            Args:
                string_to_strip (string): Te string we wish to strip
            Returns:
                string: The string with leading '0' stripped
        '''
        stripped_string = re.sub("^0+", "", string_to_strip)

        return stripped_string

    def __insert_data_to_patients_table(self, mrn, hashed_patient_mrn):
        '''
            Inserts the file's patient to the patient table
            Args:
                mrn (string): the patients mrn
                hashed_patient_mrn (stirng): the patient id
        '''
        patient_data = {
            "MRN": mrn,
            "CLIENT_ID": self.client_uuid,
            "PATIENT_ID": hashed_patient_mrn
        }

        patient_data = PatientTable.Client().standardize(patient_data)

        if Insert.Client(
                self.client_uuid).insert_data_to_patients_table(patient_data):
            return True

        Helpers.print_error('Failed to upload ' + str(patient_data) +
                            ' to patients table')
        return False

    def __create_blob_with_metadata(self, blob, file_contents, metadata):
        '''
            Creates the blob in GCS and adds metadata to it
            Args:
                blob (google.cloud.storage.blob): The blob we want to upload
                file_contents (bytes): The bytes of the file we wish to upload
                metadata (dict): The files Metadata
            Returns:
                boolean: True if the operation was successful, False if otherwise
        '''
        # add metadata to blob and save in bucket
        blob.metadata = metadata

        if not self.__try_upload_blob(blob, file_contents):
            return False

        if Helpers.try_blob_exists(blob):
            print('successfully uploaded file to datalake')
            self.__file_upload_logger.log_text('successfully uploaded file ' +
                                               blob.name + ' to datalake',
                                               severity='INFO')
            if metadata['MODALITY'] != metadata['CLASSIFIED_MODALITY']:
                self.__file_upload_logger.log_text(
                    'Classified Modality for ' + blob.name + ' (' +
                    metadata['CLASSIFIED_MODALITY'] +
                    ') does not matched metadata supplied (' +
                    metadata['MODALITY'] + ')',
                    severity='WARNING')

            return True

        return False

    def __read_file_contents(self, file_path):
        """
            Reads file from the file path, has the capabilities to read from local file,
            network file, and url. The file will be decoded to UTF-8 and remove carraige returns
            "\r" to match LF encoding.

            Args:
                file_path: The local file path, server file path or url
            Returns:
                string: string representation of data found
        """

        extension = os.path.splitext(file_path)[1].lower()

        if "gs://" in file_path:
            # GCS BUCKET
            if extension in self.text_file_list:
                content = self.__read_gcs_blob(file_path)
                return content.replace(b'\r\n', b'\n')
            else:
                file_bytes = self.__read_gcs_blob(file_path)
                return file_bytes

        # LOCAL FILE
        if extension in self.text_file_list:
            content = open(file_path, 'rb').read()
            return content.replace(b'\r\n', b'\n')
        else:
            file_bytes = open(file_path, 'rb').read()
            return file_bytes

    def __read_gcs_blob(self, gcs_blob_path):
        '''
            Reads a blob from GCS and returns the bytes 
            Args:
                gcs_blob_path (string): The path to the blob
            Returns:
                bytes: contents of the gcs blob
        '''
        file_path = gcs_blob_path.replace("gs://", "")
        bucket_name = file_path.split("/")[0]
        source_blob_name = file_path.split("/", 1)[1]
        bucket = GcpClients.instance().get_storage_client().bucket(bucket_name)
        blob = bucket.blob(source_blob_name)

        content = blob.download_as_bytes()

        return content

    def __try_upload_blob(self, blob, contents):
        '''
            Tries to upload a blob to GCS with string contents
            Args:
                blob (google.cloud.storage.Blob): The blob we wish to upload to
                contents (string): The files contents
        '''
        try:

            blob.upload_from_string(contents, content_type=blob.content_type)
            return True
        except Exception as e:
            Helpers.print_error("Could not upload blob " + str(e))
            self.__file_upload_logger.log_text(
                'Failed To Upload Blob ' + blob.name + ' using GCS client: ' +
                str(e),
                severity='ERROR')
            return False



#
#   Created By Ryan McDermott
#   Created on 6/14/2022
#

from indxdatalaketools.GoogleClients.GoogleClients import GcpClients
from google.api_core.exceptions import ServiceUnavailable
from indxdatalaketools.Helpers import print_error
from indxdatalaketools.GoogleClientWrappers.retry_decorator import retry_decorator


class Wrapper:
    '''
        Wrapper for Google Cloud Storage python SDK
    '''
    __google_cloud_tasks_client = None

    def __init__(self):
        self.__google_cloud_tasks_client = GcpClients.instance(
        ).get_tasks_v2_client()

    @retry_decorator()
    def create_http_task(self, parent, task):
        '''
            Function that creates an http task
            Args:
                - parent (string): The name of the task queue 
                    (projects/{project}/locations/{location}/queues/{queue name})
                - task (dict): Dictionary containing the task information
            Returns:
                - boolean: True if the task was created, False if otherwise
        '''
        try:
            response = self.__google_cloud_tasks_client.\
                create_task(request={"parent": parent, "task": task})
            print("Created task {}".format(response.name))
            return True
        except ServiceUnavailable as e:
            raise ServiceUnavailable('503 retry')
        except Exception as e:
            print_error(e)
            return False


#
#   Created by Ryan McDermott
#   Created on 7/7/2022
#

from indxdatalaketools.GoogleClientWrappers import SecretManager
from indxdatalaketools import Helpers

PROPERTIES_DB_URL = 'https://properties.indx.acuityeyegroup.dev/'
APPLICATION_JSON = "application/json"


class Client:
    '''
        Class that contains all the functionality make a call to Properties
        Database URL
    '''
    __project_id = ''
    __secret_id = ''
    __api_key = ''
    __properties_lists = {}

    def __init__(self):
        self.__project_id = 'indx-data-services'
        self.__secret_id = 'datalake-properties-db-api-key'
        self.__api_key = SecretManager.Wrapper().get_secret(
            self.__project_id, self.__secret_id)

    def populate_modality_lists(self):
        '''
            populates the poperties lists dictionary
        '''
        lists = {}
        headers = {
            "Content-Type": APPLICATION_JSON,
            "x-api-key": self.__api_key
        }
        response = Helpers.create_get_request(PROPERTIES_DB_URL +
                                              "property-lists",
                                              headers=headers)
        property_lists = response['data']
        for property_list in property_lists:
            attributes = property_list['attributes']

            # do not need to include non enabled property lists
            if 'enabled' in attributes and attributes['enabled']:
                lists[attributes["list-name"]] = property_list["id"]
        return lists

    def get_modalities(self):
        '''
            Function that returns a list of modlaites pulled from the properties database api
            Returns:
                - list: The list of modalities 
        '''
        if self.__properties_lists == {}:
            self.__properties_lists = self.populate_modality_lists()

        modalities = []
        headers = {
            "Content-Type": APPLICATION_JSON,
            "x-api-key": self.__api_key
        }
        modalties_id = self.__properties_lists['modalities']
        response = Helpers.create_get_request(PROPERTIES_DB_URL +
                                              "property-lists/" + modalties_id,
                                              headers=headers)
        modality_list = response['included']

        for property_value in modality_list:
            modality = property_value['attributes']["value"]
            modalities.append(modality)

        return modalities

    def get_race_codes(self):
        '''
            Function that returns the list of race codes pulled from the properties database api
            Returns:
                - list: The list of race codes
        '''
        if self.__properties_lists == {}:
            self.__properties_lists = self.populate_modality_lists()

        race_codes = []
        headers = {
            "Content-Type": APPLICATION_JSON,
            "x-api-key": self.__api_key
        }
        race_codes_id = self.__properties_lists['race-codes']
        response = Helpers.create_get_request(PROPERTIES_DB_URL +
                                              "property-lists/" + race_codes_id,
                                              headers=headers)
        modality_list = response['included']

        for property_value in modality_list:
            race_code = property_value['attributes']["value"]
            race_codes.append(race_code)

        return race_codes



#
#   Created by Ryan McDermott
#   Created on 2/21/2022
#

import datetime
import time

from indxdatalaketools.ClientTools.PatientsTable import Update
from indxdatalaketools import Helpers
from indxdatalaketools.GoogleClientWrappers import BigQuery
from indxdatalaketools.GoogleClients.GoogleClients import GcpClients


class Client:
    '''
        Class that contains all the functionality to insert a patient into
        The patient table
    '''

    client_uuid = ''
    dataset_id = ''
    patient_table_name = ''
    table_fields = [
        "CLIENT_ID", "MRN", "PATIENT_ID", "FIRSTNAME", "MIDDLENAME", "LASTNAME",
        "DOB", "SEX", "RACE", 'CHECKSUM', 'UPDATED_AT'
    ]
    __patient_insert_logger = None

    def __init__(self, client_uuid):
        '''
            Init function that sets up client credentials, uuid, 
            and google api clients used to insert items to the Patients table
            Args:
                google_clients          (ApiClients): object containing all Google Api Clients
                client_uuid             (string): The client's uuid
            Returns:
                None
        '''
        self.client_uuid = client_uuid
        self.dataset_id = self.client_uuid.replace("-", "_")
        self.patient_table_name                 = 'indx-data-services.' + \
            self.dataset_id + ".PATIENTS"
        self._bq_wrapper = BigQuery.Wrapper()
        logging_client = GcpClients.instance().get_logging_client()
        self.__patient_insert_logger = logging_client.logger(client_uuid +
                                                             '-patient-logger')

    def insert_data_to_patients_table(self, patient_json, overwrite=False):
        '''
            This function determines whether the patient_json is a list
            or a single patient. Then it uploads all patient data to the PATIENT 
            table for the client_id
            Args:
                patient_json    (json): Either a json object or a JSON Array containing
                    multiple patients data
                overwrite       (boolean): Flag to indicat whether or not we can overwrite the patients 
                    data
            Returns: 
                boolean: True if the patient/s were succesfully inserted in the PATIENTS table
        '''
        # parse json to determine single or multiple patients

        # multiple patients
        if isinstance(patient_json, list):
            return self.__insert_multiple_patients_to_table(patient_json,
                                                            overwrite=overwrite)

        return self.__insert_patient_to_patients_table(patient_json,
                                                       overwrite=overwrite)

    def __insert_multiple_patients_to_table(self,
                                            json_structure,
                                            overwrite=False):
        '''
            loops through json structure and inserts all patients into the PATIENTS
            table, if a patient is not successfully uploaded then it is logged and
            the process continues.
            Args:
                json_structure (dict[]): list of dictionarys containing the patients data
                overwrite (boolean): Flag to indicate whether or not the row needs to get
                    overwritten
            Returns:
                boolean: True if the operation was a success False if otherwise
        '''
        status = True
        for i in json_structure:
            if self.__insert_patient_to_patients_table(i, overwrite=overwrite):
                continue
            else:
                status = False

        if status == False:
            self.__patient_insert_logger.log_text(
                'One or more patients have not been uploaded Successfully',
                severity='ERROR')

        self.__patient_insert_logger.log_text(
            'All patients have been uploaded Successfully', severity='INFO')
        return status

    def __insert_patient_to_patients_table(self, patient_data, overwrite=False):
        '''
            Function that adds a single patient to the patient table, if that patient already exists
            in the table their data is updated with the provided information
            Args:
                patient_data    (dict): The data we wish to add to the patient table
                overwrite       (boolean): Flag to indicate whether or not we can over 
                    write the patients data
            Returns:
                boolean: True if the operation was a success, false if otherwise
        '''
        now = datetime.datetime.utcnow()
        patient_data['UPDATED_AT'] = str(now).replace(' ', 'T')
        patient_mrn = patient_data['MRN']
        patient_partition = patient_data['PATIENT_PARTITION']
        patient_id = Helpers.patient_hash(self.client_uuid, [patient_mrn])[0]
        start = time.time()
        patient_row = self.__get_patient_in_patients_table(
            patient_mrn, patient_id, patient_partition)
        finish = time.time()
        print('time to get patient: ' + str(finish - start) + ' seconds')
        start = time.time()
        result = self.__insert_data_to_patients_table(patient_data,
                                                      patient_row,
                                                      overwrite=overwrite)
        finish = time.time()
        print('time to upload patient data: ' + str(finish - start) +
              ' seconds')
        return result

    def __get_patient_in_patients_table(self, patient_mrn, patient_id,
                                        patient_partition):
        '''
            performs a query look up on the PATIENTS table to find if the given patient
            exists in the PATIENTS table
            Args:
                patient_mrn (string): The patients mrn
                patient_id  (string): The patients UUID
            Returns:
                The row of the Patient, if they are not in the table None is returned
        '''
        query = 'SELECT * \
                    FROM `' + self.patient_table_name + '`\
                    WHERE MRN =\'' + patient_mrn + '\'\
                    AND PATIENT_ID =\'' + patient_id + '\'\
                    AND PATIENT_PARTITION =' + str(patient_partition)

        rows = BigQuery.Wrapper().perform_query(query)

        if rows is None:
            return None

        if rows.total_rows == 0:
            return None
        if rows.total_rows > 1:
            Helpers.print_error(
                "CRITICAL: Error in PATIENTS table, multiple instances of MRN: "
                + patient_mrn + " and PATIENT_ID " + patient_id)

        # must only contain one row if not 0 and > 1
        return next(rows)

    def __insert_data_to_patients_table(self,
                                        patient_data,
                                        patient_row,
                                        overwrite=False):
        '''
            Inserts data to the patients table, if the data needs to be updated it will overwrite
            if the flag is set
            Args:
                patient_data (dict): The dictionary of the patients data
                patient_row  (row): The row already existing in the patients table
                overwrite    (boolean): True if we wish to overwrite the existing data in the 
                    PATIENTS table
            Returns:
                boolean: True if the operations was successful, False if otherwise
        '''
        if patient_row is not None:
            # patient exists, updated patient in table
            return Update.Client(
                self.client_uuid).update_patient_in_patient_table(
                    patient_data, patient_row, overwrite=overwrite)
        elif overwrite == False:
            # if not append to PATIENTS table
            return self.__add_data_to_patient_table(patient_data)
        else:
            Helpers.print_error("patient does not exist cannot update")
            return False

    def __add_data_to_patient_table(self, patient_data):
        '''
            Adds the patients data to the PATIENT table using the Big Query 
            Storage API COMMITTED row type to bypass the streaming buffer
            Args:
                patient_data    (dict): The Patients Data
            Returns
                Boolean: True if the request worked, false if other wise
        '''
        result = self._bq_wrapper.insert_to_bq_table(
            [patient_data],
            'indx-data-services.' + self.dataset_id + '.PATIENTS')

        if result == True:
            self.__patient_insert_logger.log_text(
                str(patient_data) + ' has been succesfully uploaded',
                severity='INFO')
        else:
            self.__patient_insert_logger.log_text(str(patient_data) +
                                                  ' could not be uploaded',
                                                  severity='ERROR')
        return result



from importlib.metadata import version, PackageNotFoundError
# pre-3.8 import statement
# from importlib_metadata import version, PackageNotFoundError
VERSION_FALLBACK = "0.0.1"
__version__ = ''
try:
    __version__ = version('indxdatalaketools')
except PackageNotFoundError:
    # package is not installed
    # assign signal or sane value as a default
    __version__ = VERSION_FALLBACK
    pass


#
#   Created by Ryan McDermott
#   Created on 3/16/2022
#

from google.api_core.exceptions import NotFound
from google.api_core.exceptions import ServiceUnavailable
from indxdatalaketools import Helpers
from indxdatalaketools.GoogleClientWrappers.retry_decorator import retry_decorator
from indxdatalaketools.GoogleClients.GoogleClients import GcpClients


class Wrapper:
    ''' Wrapper for all scheduler client API Calls'''
    __scheduler_client = None

    def __init__(self):
        self.__scheduler_client = GcpClients.instance().get_scheduler_v1_client(
        )

    @retry_decorator()
    def check_job_exists(self, job):
        '''
            Chekcs if the job already exists
            Args:
                job (dict): Dictionary of job request
            Returns:
                boolean: True if the job exists, False if otherwise
        '''
        try:
            self.__scheduler_client.get_job(request=job)
            return True
        except ServiceUnavailable:
            raise ServiceUnavailable('503 retry')
        except NotFound:
            return False

    @retry_decorator()
    def try_create_job(self, parent, job):
        ''' 
            Tries to create a cloud scheduler job
            Args:
                parent (google.cloud.scheduler_v1.types.Parent): The parent of the job
                job (google.cloud.scheduler_v1.types.Job): The job we wish to create
            Returns:
                boolean: True if the job was created False if otherwise
        '''
        try:
            self.__scheduler_client.create_job(request={
                "parent": parent,
                "job": job
            })
            print('Scheduled job was created')
            return True
        except ServiceUnavailable:
            raise ServiceUnavailable('503 retry')
        except Exception as exception:
            Helpers.print_error('Could not create cloud scheduler job: ' +
                                str(exception))
            return False





import datetime
import re
from indxdatalaketools import Helpers

from indxdatalaketools.DataLakeGlobals import DATE_REGEX


def validate_date(date):
        '''
            Validates the date
            Args:
                date (string): The date
            Returns:
                boolean: True if the date is valid, false if otherwise
        '''
        if re.fullmatch(DATE_REGEX, date) is None:
            print(date + ' does not match date regex: not a valid date')
            return False
        
        # check if valid date
        year    = int(date[:4])
        month   = int(date[4:6])
        day     = int(date[6:8])
        
        if __is_valid_date(year, month, day):
            return True

        Helpers.print_error(date + ' is not a valid date')
        return False

    
def __is_valid_date( year, month, day):
    '''
        checks if the year month and day creates a valid date
        Args:
            year    (string): The year
            month   (string): The month
            day     (string): the day
        Returns:
            boolean: True if a valid date is formed False if other wise
    '''
    try:
        datetime.datetime(year, month, day)
        return True
    except:
        return False





#
#   Created by Ryan McDermott
#   Created on 3/2/2022
#   Updated on 9/30/2022: Changed class to singleton
#

from __future__ import annotations
import json

import google.auth

from google.cloud import logging
from google.cloud import storage
from google.cloud import bigquery
from google.cloud import bigquery_storage_v1
from google.cloud import datastore
from google.cloud import pubsub_v1
from google.cloud import scheduler_v1
from google.cloud import tasks_v2
from google.cloud import secretmanager

from google.oauth2 import service_account
from google.cloud.compute_v1.services import instances
from google.cloud.compute_v1.services import zone_operations

import os.path


class GcpClients:
    """A singleton class to hold all gcs clients.

    This class hold all gcp clients to use throughout your code. 
    Pass in credentials upon first use of the GcpClients class and they will be saved. If
    a certain gcp client has not been initialized on its first "get_..." it will be initialized.
        

    Typical usage example:

        gcs_client = GcpClients.instance().get_gcs_client()
    """
    _credentials = None
    _instance = None
    _logging_client = None
    _storage_client = None
    _bigquery_client = None
    _bigquery_storage_v1_client = None
    _datastore_client = None
    _secretmanager_client = None
    _pubsub_v1_client = None
    _tasks_v2_client = None
    _scheduler_v1_client = None
    _compute_engine_instance_client = None
    _compute_engine_operation_client = None
    _account_credentials_client = None

    def __init__(self):
        '''This will throw a RunTimeError.

        Will never be able to instanciate this object.
        call GcpClients.instance() instead

        Raises:
            RuntimeError: Always raises runtime error
        '''
        raise RuntimeError('Call instance() instead')

    @classmethod
    def instance(cls, credential_file: str | None = None) -> GcpClients:
        '''singleton design pattern, call this function to get the singleton

        Static method that returns the instance of the GcpClients object. If the instance is none,
        a new GcpClients will be created. This is the only way to use the object.

        Args:
            credential_file (str, optional): The credentials used to 
                authenticate to GCP, defaults to None
        Returns:
            GcpClients: The GcpClients singleton
        Raises:
            FileNotFoundError: if the passed credentials do not exist
            ValueError: if the passed credentials file is not a json

        '''
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
            if credential_file is not None and not os.path.isfile(
                    credential_file):
                raise FileNotFoundError(
                    "The Credential file path does not exist!")
            if credential_file is not None:
                # will raise ValueError if the credential file is not json
                json.load(open(credential_file))

            cls._instance._credentials = credential_file

        return cls._instance

    def get_credentials(self) -> str | None:
        ''' Gets the current credentials

        Returns:
            str | None: the credentials passed to the GcpClients Singleton
        '''
        return self._credentials

    def get_logging_client(self) -> logging.Client:
        '''Gets the authenticated GCP logging client.

        If the current logging client has not been initialized then one will be created 
        with the credentials in the GcpClient.

        Returns:
            google.cloud.logging.Client: The authenticated logging client
        '''
        if self._logging_client is None:
            if self._credentials is not None:
                self._logging_client = logging.Client.\
                from_service_account_json(self._credentials)

            else:
                self._logging_client = logging.Client()
        return self._logging_client

    def get_storage_client(self) -> storage.Client:
        '''Gets the authenticated GCP storage client.

        If the current storage client has not been initialized then one will be created 
        with the credentials in the GcpClient.

        Returns:
            google.cloud.storage.Client: The authenticated storage client
        '''
        if self._storage_client is None:
            if self._credentials is not None:
                self._storage_client = storage.Client.\
                from_service_account_json(self._credentials)
            else:
                self._storage_client = storage.Client()
        return self._storage_client

    def get_bigquery_client(self) -> bigquery.Client:
        '''Gets the authenticated GCP bigquery client.

        If the current bigquery client has not been initialized then one will be created 
        with the credentials in the GcpClient.

        Returns:
            google.cloud.bigquery.Client: The authenticated bigquery client
        '''
        if self._bigquery_client is None:
            if self._credentials is not None:
                self._bigquery_client = bigquery.Client.\
                from_service_account_json(self._credentials)
            else:
                self._bigquery_client = bigquery.Client()
        return self._bigquery_client

    def get_bigquery_storage_v1_client(
            self) -> bigquery_storage_v1.BigQueryWriteClientClient:
        '''Gets the authenticated GCP bigquery_storage_v1 Write client.

        If the current bigquery_storage_v1 Write client has not 
        been initialized then one will be created with the credentials in the GcpClient.

        Returns:
            google.cloud.bigquery_storage_v1.BigQueryWriteClientClient: 
                The authenticated bigquery client
        '''
        if self._bigquery_storage_v1_client is None:
            if self._credentials is not None:
                self._bigquery_storage_v1_client = bigquery_storage_v1.BigQueryWriteClient.\
                from_service_account_json(self._credentials)
            else:
                self._bigquery_storage_v1_client = bigquery_storage_v1.BigQueryWriteClient(
                )
        return self._bigquery_storage_v1_client

    def get_datastore_client(self) -> datastore.Client:
        '''Gets the authenticated GCP datastore client.

        If the current datastore client has not been initialized then one will be created 
        with the credentials in the GcpClient.

        Returns:
            datastore.Client: The authenticated datastore client
        '''
        if self._datastore_client is None:
            if self._credentials is not None:
                self._datastore_client = datastore.Client.\
                from_service_account_json(self._credentials)
            else:
                self._datastore_client = datastore.Client()

        return self._datastore_client

    def get_secretmanager_client(
            self) -> secretmanager.SecretManagerServiceClient:
        '''Gets the authenticated GCP secretmanager client.

        If the current secretmanager client has not been initialized then one will be created 
        with the credentials in the GcpClient.

        Returns:
            secretmanager.SecretManagerServiceClient: The authenticated secretmanager client
        '''
        if self._secretmanager_client is None:
            if self._credentials is not None:
                self._secretmanager_client = secretmanager.SecretManagerServiceClient.\
                from_service_account_json(self._credentials)
            else:
                self._secretmanager_client = secretmanager.SecretManagerServiceClient(
                )

        return self._secretmanager_client

    def get_pubsub_v1_client(self) -> pubsub_v1.Client:
        '''Gets the authenticated GCP pubsub_v1 client.

        If the current pubsub_v1 client has not been initialized then one will be created 
        with the credentials in the GcpClient.

        Returns:
            pubsub_v1.Client: The authenticated pubsub_v1 client
        '''
        if self._pubsub_v1_client is None:
            if self._credentials is not None:
                self._pubsub_v1_client = pubsub_v1.PublisherClient.\
                from_service_account_json(self._credentials)
            else:
                self._pubsub_v1_client = pubsub_v1.PublisherClient()

        return self._pubsub_v1_client

    def get_tasks_v2_client(self) -> tasks_v2.CloudTasksClient:
        '''Gets the authenticated GCP tasks_v2 client.

        If the current tasks_v2 client has not been initialized then one will be created 
        with the credentials in the GcpClient.

        Returns:
            scheduler_v1.CloudSchedulerClient: The authenticated tasks_v2 client
        '''
        if self._tasks_v2_client is None:
            if self._credentials is not None:
                self._tasks_v2_client = tasks_v2.CloudTasksClient.\
                from_service_account_json(self._credentials)
            else:
                self._tasks_v2_client = tasks_v2.CloudTasksClient()

        return self._tasks_v2_client

    def get_scheduler_v1_client(self) -> scheduler_v1.CloudSchedulerClient:
        '''Gets the authenticated GCP scheduler_v1 client.

        If the current scheduler_v1 client has not been initialized then one will be created 
        with the credentials in the GcpClient.

        Returns:
            scheduler_v1.CloudSchedulerClient: The authenticated scheduler_v1 client
        '''
        if self._scheduler_v1_client is None:
            if self._credentials is not None:
                self._scheduler_v1_client = scheduler_v1.CloudSchedulerClient.\
                from_service_account_json(self._credentials)
            else:
                self._scheduler_v1_client = scheduler_v1.CloudSchedulerClient()

        return self._scheduler_v1_client

    def get_compute_engine_instance_client(self) -> instances.InstancesClient:
        '''Gets the authenticated GCP compute engine instance client.

        If the current compute engine instance client has not been initialized 
        then one will be created with the credentials in the GcpClient.

        Returns:
            instances.InstancesClient: The authenticated compute engine instance client
        '''
        if self._compute_engine_instance_client is None:
            if self._credentials is not None:
                self._compute_engine_instance_client = instances.InstancesClient.\
                from_service_account_json(self._credentials)
            else:
                self._compute_engine_instance_client = instances.InstancesClient(
                )

        return self._compute_engine_instance_client

    def get_compute_engine_operation_client(
            self) -> zone_operations.ZoneOperationsClient:
        '''Gets the authenticated GCP compute engine operation client.

        If the current compute engine operation client has not been initialized 
        then one will be created with the credentials in the GcpClient.

        Returns:
            zone_operations.ZoneOperationsClient: The authenticated compute engine operation client
        '''
        if self._compute_engine_operation_client is None:
            if self._credentials is not None:
                self._compute_engine_operation_client = zone_operations.ZoneOperationsClient.\
                from_service_account_json(self._credentials)
            else:
                self._compute_engine_operation_client = zone_operations.ZoneOperationsClient(
                )

        return self._compute_engine_operation_client

    def get_account_credentials_client(self) -> service_account.Credentials:
        '''Gets the authenticated GCP service account credentials.

        If the current service account credentials has not been initialized 
        then one will be created with the credentials in the GcpClient.

        Returns:
            service_account.Credentials: The authenticated service account credentials
        '''
        if self._account_credentials_client is None:
            if self._credentials is not None:
                self._account_credentials_client = service_account.Credentials.\
                from_service_account_file(filename=self._credentials,
                scopes=['https://www.googleapis.com/auth/cloud-platform'])
            else:
                account_cred, _ = google.auth.default()
                self._account_credentials_client = account_cred

        return self._account_credentials_client



import time
from google.api_core.exceptions import ServiceUnavailable


def retry_decorator(retries: int = 5, backoff_in_seconds: int = 1):
    ''' Decorator to retry a function on a 503 ServiceUnavailable exception

    This retry decorator uses exponential back off with a default value of 5 retries
    and 1 second for backoff

    Args:
        func (function): the function we wish to use the retry strategy on
        retries (int): optional paramter with the number of retries for the client call
        backoff_in_seconds (int): optional parameter the number of seconds to wait initially

    Returns:
        function: the wrapped function
    '''

    def rd(func):

        def retry(*args, **kwargs):
            for x in range(0, retries):
                try:
                    returned_value = func(*args, **kwargs)
                    break
                except ServiceUnavailable:
                    # exponential backoff
                    sleep = (backoff_in_seconds * 2**x)
                    time.sleep(sleep)

            return returned_value

        return retry

    return rd
